"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.optionsDefault = exports.OnesyHuffmanTree = exports.OnesyNode = exports.OnesyHuffmanCodeResponse = void 0;
const is_1 = __importDefault(require("@onesy/utils/is"));
const merge_1 = __importDefault(require("@onesy/utils/merge"));
const copy_1 = __importDefault(require("@onesy/utils/copy"));
const to_1 = __importDefault(require("@onesy/utils/to"));
const binaryStringToHexadecimal_1 = __importDefault(require("@onesy/utils/binaryStringToHexadecimal"));
const hexadecimalStringToBinary_1 = __importDefault(require("@onesy/utils/hexadecimalStringToBinary"));
const OnesyDate_1 = __importDefault(require("@onesy/date/OnesyDate"));
const duration_1 = __importDefault(require("@onesy/date/duration"));
class OnesyHuffmanCodeResponse {
    constructor(value = '', values, values_encoded, probabilities, efficiency, redundency, entropy, original_byte_size, values_byte_size, value_byte_size, encoded_byte_size, compression_ratio, compression_percentage, positive, average_code_word_length, performance_milliseconds, performance) {
        this.value = value;
        this.values = values;
        this.values_encoded = values_encoded;
        this.probabilities = probabilities;
        this.efficiency = efficiency;
        this.redundency = redundency;
        this.entropy = entropy;
        this.original_byte_size = original_byte_size;
        this.values_byte_size = values_byte_size;
        this.value_byte_size = value_byte_size;
        this.encoded_byte_size = encoded_byte_size;
        this.compression_ratio = compression_ratio;
        this.compression_percentage = compression_percentage;
        this.positive = positive;
        this.average_code_word_length = average_code_word_length;
        this.performance_milliseconds = performance_milliseconds;
        this.performance = performance;
    }
}
exports.OnesyHuffmanCodeResponse = OnesyHuffmanCodeResponse;
class OnesyNode {
    constructor(value = 1, word) {
        this.value = value;
        this.word = word;
        this.value = Number(value.toFixed(3));
    }
    get leaf() {
        return !(this.left || this.right);
    }
    get maxDepth() {
        const maxDepthMethod = (value) => {
            if (value === undefined)
                return 0;
            return Math.max(1 + maxDepthMethod(value.left), 1 + maxDepthMethod(value.right));
        };
        return maxDepthMethod(this);
    }
}
exports.OnesyNode = OnesyNode;
class OnesyHuffmanTree {
    static make(value) { return new OnesyHuffmanTree().make(value); }
    get array() {
        const value = [];
        this.preorder(this.root, (value_) => {
            var _a;
            value.push(value_.word ? value_.word : value_ === this.root ? 0 : (_a = value_.path) === null || _a === void 0 ? void 0 : _a.slice(-1));
        });
        return value;
    }
    isRoot(value) {
        return value === this.root;
    }
    preorder(value, method) {
        if (value !== undefined && (0, is_1.default)('function', method)) {
            method(value, value.left, value.right);
            this.preorder(value.left, method);
            this.preorder(value.right, method);
        }
    }
    make(value_) {
        const items = (0, copy_1.default)(value_);
        const onesyHuffmanTree = new OnesyHuffmanTree();
        onesyHuffmanTree.root = new OnesyNode();
        onesyHuffmanTree.root.index = 0;
        function arrayToOnesyHuffmanTree(value) {
            if (items[0] === '0' && !value.left) {
                value.left = new OnesyNode();
                value.left.index = 2 * value.index + 1;
                value.left.path = '0';
                items.splice(0, 1);
                arrayToOnesyHuffmanTree(value.left);
            }
            if (items[0] === '1' && !value.right) {
                value.right = new OnesyNode();
                value.right.index = 2 * value.index + 2;
                value.right.path = '1';
                items.splice(0, 1);
                arrayToOnesyHuffmanTree(value.right);
            }
            if ((0, is_1.default)('array', items[0])) {
                if (items[0].length) {
                    if (!value.left) {
                        value.left = new OnesyNode(1, items[0][0]);
                        value.left.index = 2 * value.index + 1;
                        value.left.path = '0';
                        items[0].splice(0, 1);
                    }
                    if (!value.right && items[0].length) {
                        value.right = new OnesyNode(1, items[0][0]);
                        value.right.index = 2 * value.index + 2;
                        value.right.path = '1';
                        items[0].splice(0, 1);
                    }
                    if (!items[0].length)
                        items.splice(0, 1);
                }
                else
                    items.splice(0, 1);
            }
            if (items[0] === '1' && !value.right)
                arrayToOnesyHuffmanTree(value);
        }
        arrayToOnesyHuffmanTree(onesyHuffmanTree.root);
        return onesyHuffmanTree;
    }
}
exports.OnesyHuffmanTree = OnesyHuffmanTree;
exports.optionsDefault = {
    encode_values: true,
    base64: true
};
class OnesyHuffmanCode {
    constructor(value, options = exports.optionsDefault) {
        this.value = value;
        this.probabilities = {};
        this.values = {};
        this.response = new OnesyHuffmanCodeResponse();
        this.options = (0, merge_1.default)(options, exports.optionsDefault);
        if (this.value !== undefined)
            this.init();
    }
    static get OnesyHuffmanCodeResponse() { return OnesyHuffmanCodeResponse; }
    static get OnesyNode() { return OnesyNode; }
    static get OnesyHuffmanTree() { return OnesyHuffmanTree; }
    static encodeValue(value) {
        if (!((0, is_1.default)('string', value) && value.length))
            return '';
        // Add 1 at the start of every 3 characters
        // it's more data, but there will be no bugs
        // with padded 0s, a bug fix for now
        return (0, binaryStringToHexadecimal_1.default)((value.match(/.{1,3}/g) || []).map(item => 1 + item).join('')).match(/.{1,2}/g).flatMap(item => {
            if (item[0] === '0')
                return item.split('').map(item_ => String.fromCharCode(parseInt(item_, 16)));
            return String.fromCharCode(parseInt(item, 16));
        }).join('');
    }
    static decodeValue(value_) {
        if (!((0, is_1.default)('string', value_) && value_.length))
            return '';
        const value = value_.split('').map(item => item.charCodeAt(0).toString(16)).join('');
        return ((0, hexadecimalStringToBinary_1.default)(value).match(/.{1,4}/g) || []).map(item => item.slice(1)).join('');
    }
    static encodeValues(values, encodeValues = true) {
        if (values) {
            let result = '';
            const keys = Object.keys(values);
            keys.forEach((item, index) => result += `${item}${encodeValues ? this.encodeValue(values[item]) : values[item]}${index < keys.length - 1 ? ' ' : ''}`);
            return `${encodeValues ? 1 : 0}${result}`;
        }
    }
    static decodeValues(value) {
        const result = {};
        const values = [];
        if (value) {
            const encodeValues = value[0] === '1';
            const values_ = value.slice(1).split(' ');
            values_.forEach((item, index) => {
                if (!item)
                    values_[index + 1] = ` ${values_[index + 1]}`;
                else
                    values.push(item);
            });
            values.forEach(item => result[item[0]] = encodeValues ? this.decodeValue(item.slice(1)) : item.slice(1));
        }
        return result;
    }
    static getValues(onesyHuffmanTree) {
        const values = {};
        const leafs = [];
        if (onesyHuffmanTree) {
            onesyHuffmanTree.preorder(onesyHuffmanTree.root, (value, left, right) => {
                if (onesyHuffmanTree.isRoot(value)) {
                    value.path = value.maxDepth === 1 ? '0' : '';
                    if (value.leaf)
                        leafs.push(value);
                }
                if (left) {
                    left.path = value.path + 0;
                    if (left.leaf)
                        leafs.push(left);
                }
                if (right) {
                    right.path = value.path + 1;
                    if (right.leaf)
                        leafs.push(right);
                }
            });
        }
        leafs.filter(leaf => leaf.word).forEach(leaf => values[leaf.word] = leaf.path);
        return values;
    }
    static decode(value, values) {
        const instance = new OnesyHuffmanCode();
        instance.values = values;
        return instance.decode(value);
    }
    static encodeBase64(value) {
        return (0, to_1.default)(value, 'base64');
    }
    static decodeBase64(value) {
        return (0, to_1.default)(value, 'string');
    }
    get encoded() {
        return this.response;
    }
    get entropy() {
        const output = Object.keys(this.probabilities).reduce((result, key) => result += this.probabilities[key] * (Math.log2(this.probabilities[key])), 0);
        return Math.abs(Number(output.toFixed(3)));
    }
    get averageCodeWordLength() {
        const output = Object.keys(this.probabilities).reduce((result, key) => { var _a; return result += this.probabilities[key] * (((_a = this.values[key]) === null || _a === void 0 ? void 0 : _a.length) || 8); }, 0);
        return Number((output).toFixed(3));
    }
    get redundency() {
        return Number(Math.abs(this.entropy - this.averageCodeWordLength).toFixed(3));
    }
    get efficiency() {
        return Number(((this.entropy / this.averageCodeWordLength) || 0).toFixed(3));
    }
    init() {
        this.startTime = OnesyDate_1.default.milliseconds;
        if (!Object.keys(this.probabilities).length && (0, is_1.default)('string', this.value)) {
            // Frequencies
            this.getProbabilities();
        }
        if (!Object.keys(this.values).length && Object.keys(this.probabilities).length) {
            // Normalize probabilities
            this.normalizeProbabilities();
            // Make huffman tree
            this.makeHuffmanTree();
            // Values
            this.values = OnesyHuffmanCode.getValues(this.huffmanTree);
        }
        // Encode
        this.encode();
    }
    encode() {
        const response = new OnesyHuffmanCodeResponse();
        if (Object.keys(this.values).length && (0, is_1.default)('string', this.value)) {
            let value = Array.from(this.value).reduce((result, item) => result += (this.values[item] || item), '');
            value = OnesyHuffmanCode.encodeValue(value);
            if (this.options.base64)
                value = OnesyHuffmanCode.encodeBase64(value);
            response.value = value;
            response.performance_milliseconds = OnesyDate_1.default.milliseconds - this.startTime;
            response.performance = (0, duration_1.default)(response.performance_milliseconds) || '0 milliseconds';
            response.values = this.values;
            response.values_encoded = OnesyHuffmanCode.encodeValues(this.values, this.options.encode_values);
            response.probabilities = this.probabilities;
            response.efficiency = this.efficiency;
            response.redundency = this.redundency;
            response.entropy = this.entropy;
            response.average_code_word_length = this.averageCodeWordLength;
            response.original_byte_size = (0, to_1.default)(this.value, 'byte-size');
            response.values_byte_size = (0, to_1.default)(response.values_encoded, 'byte-size');
            response.value_byte_size = (0, to_1.default)(value, 'byte-size');
            response.encoded_byte_size = response.values_byte_size + response.value_byte_size;
            response.compression_ratio = Number((((response.encoded_byte_size + response.original_byte_size) / response.encoded_byte_size) - 1).toFixed(2));
            response.compression_percentage = response.original_byte_size === 0 ? response.value_byte_size === 0 ? 0 : response.value_byte_size * -100 : Number((((response.original_byte_size - response.encoded_byte_size) / response.original_byte_size) * 100).toFixed(2));
            response.positive = response.compression_ratio > 1;
            this.response = response;
        }
        return response;
    }
    decode(value_) {
        if (!value_)
            return new OnesyHuffmanCodeResponse(value_);
        const response = new OnesyHuffmanCodeResponse(value_);
        const startTime = OnesyDate_1.default.milliseconds;
        const value = OnesyHuffmanCode.decodeValue(OnesyHuffmanCode.decodeBase64(value_));
        if ((0, is_1.default)('string', value) && Object.keys(this.values).length) {
            let input = value;
            let output = '';
            while (input.length) {
                let valueWord = Object.keys(this.values).find(key => input.indexOf(this.values[key]) === 0);
                if (!valueWord) {
                    // bug
                    valueWord = (Object.keys(this.values).find(key => ('0' + input).indexOf(this.values[key]) === 0) ||
                        Object.keys(this.values).find(key => ('00' + input).indexOf(this.values[key]) === 0));
                    if (!valueWord)
                        break;
                }
                output += valueWord;
                input = input.slice(this.values[valueWord].length);
            }
            response.value = output;
            response.performance_milliseconds = OnesyDate_1.default.milliseconds - startTime;
            response.performance = (0, duration_1.default)(response.performance_milliseconds) || '0 milliseconds';
            response.original_byte_size = (0, to_1.default)(output, 'byte-size');
            response.value_byte_size = (0, to_1.default)(value_, 'byte-size');
        }
        return response;
    }
    getProbabilities() {
        const value = this.value || '';
        for (let i = 0; i < value.length; i++) {
            this.probabilities[value[i]] = ~~this.probabilities[value[i]] + 1;
        }
        return this.probabilities;
    }
    normalizeProbabilities() {
        const sum = Object.keys(this.probabilities).reduce((result, item) => result += this.probabilities[item], 0);
        Object.keys(this.probabilities).forEach(key => this.probabilities[key] = Number((this.probabilities[key] / sum).toFixed(4)));
        return this.probabilities;
    }
    makeHuffmanTree() {
        let trees = [];
        Object.keys(this.probabilities).forEach(key => {
            const onesyNode = new OnesyNode(this.probabilities[key], key);
            trees.push(onesyNode);
        });
        trees.sort((a, b) => a.value - b.value);
        while (trees.length > 1) {
            const first = trees[0];
            const second = trees[1];
            const newNode = new OnesyNode(first.value + second.value);
            const children = [first, second].sort((a, b) => {
                const aMaxDepth = a.maxDepth;
                const bMaxDepth = b.maxDepth;
                if ((a.leaf && b.leaf) ||
                    (aMaxDepth === b.maxDepth))
                    return b.value - a.value;
                if (a.leaf || b.leaf)
                    return a.leaf ? -1 : 1;
                return aMaxDepth - bMaxDepth;
            });
            newNode.left = children[0];
            newNode.right = children[1];
            trees.push(newNode);
            trees = trees.slice(2);
            trees.sort((a, b) => a.value - b.value);
        }
        this.huffmanTree = new OnesyHuffmanTree();
        this.huffmanTree.root = trees[0];
        return this.huffmanTree;
    }
}
exports.default = OnesyHuffmanCode;
